package ch.hei.kart;

import java.util.EnumSet;
import java.util.LinkedList;
import java.util.List;
import android.os.Handler;

import ch.hei.kart.comm.CommunicationSetup;
import ch.hei.kart.comm.CommunicationSetupListener;
import ch.hei.kart.comm.discovery.DiscoveryEntity;
import ch.hei.kart.comm.protocol.Protocol;
import ch.hei.kart.comm.protocol.ProtocolListener;
import ch.hei.kart.comm.protocol.ProtocolState;

/**
 * Kart model. Once connected, the kart model allows to manipulate the kart developed during the 
 * first year's summer school remotely.
 * 
 * @author Nils Chatton (chn@hevs.ch)
 * @author Christopher Metrailler (mei@hevs.ch)
 * @author Michael Clausen (clm@hevs.ch)
 * 
 * @version 2.0 - 2014
 */
public class Kart {
	/**
	 * Maximal drive speed.
	 */
	public static final int MAXIMAL_SPEED			=  15;
	
	/**
	 * Minimal drive speed (maximal backwards).
	 */
	public static final int MINIMAL_SPEED			= -15;
	
	// Battery ADC has this value when batteries are 100% charged.
	// 1V ~ 200 => Fully charged (>13V) ~ 26000, empty (<11V) ~ 22000.
	private static final float MAXIMAL_BATT_LEVEL	= 26000f;
	private static final float MINIMAL_BATT_LEVEL	= 22000f;
	
	
	/*** Construction *****************************************************************************/
	
	/**
	 * Constructs a new unconnected kart object with the default setup values.
	 */
	public Kart() {
		// Create a kart object with the default setup.
		this(new KartSetup());
	}
	
	/**
	 * Constructs a new kart object with the given setup.
	 * 
	 * @param setup	Kart setup to apply.
	 */
	public Kart(KartSetup setup) {
		// Save setup and subscribe to setup changes.
		this.setup = setup;
		setup.setListener(setupListener);
		
		// Copy the setup values to the corresponding registers.
		setControlRegister(KartControlRegister.DrivePwmPeriod, setup.drivePwmPeriod());
		setControlRegister(KartControlRegister.SteeringStepPeriod, setup.steeringStepPeriod());
		setControlRegister(KartControlRegister.SteeringEndSwitchAddress, 
					setup.steeringEndSwitchAddress());
		setControlRegister(KartControlRegister.HardwareSettings, 
					KartHardwareSettings.toInt(setup.hardwareSettings()));
		setControlRegister(KartControlRegister.UpdateInterval, setup.updateIntervalMs());
		
		// Initialize other registers.
		setControlRegister(KartControlRegister.DriveSpeed, 0);
		setControlRegister(KartControlRegister.SteeringPosition, setup.steeringCenterPosition());
		setControlRegister(KartControlRegister.LEDs, 0);
		
		// Make all registers dirty in order to transmit them ASAP.
		uncommittedControlRegisters = EnumSet.allOf(KartControlRegister.class);
		
		// Register the main kart status register listener in order to forward the events to the 
		// simpler kart listeners.
		addStatusRegisterListener(kartListenerDispatcher);
	}

	
	/*** Kart setup *******************************************************************************/
	
	/**
	 * Returns a reference to the kart setup. Note that when you change settings on the 
	 * 		  kart setup object, these are applied immediately to the kart.
	 * 
	 * @return	Kart setup object for the kart instance.
	 */
	public KartSetup setup() {
		return setup;
	}

	// The actual kart setup object.
	private KartSetup setup;
	
	// Kart setup listener processing setup changes.
	private KartSetupListener setupListener = new KartSetupListener() {
		@Override
		public void parameterChanged(SetupParameter parameter) {
			switch (parameter) {
				case DrivePwmPeriod:
					setControlRegister(KartControlRegister.DrivePwmPeriod, 
									   setup.drivePwmPeriod());
					break;
					
				case SteeringStepPeriod:
					setControlRegister(KartControlRegister.SteeringStepPeriod, 
								setup.steeringStepPeriod());
					break;

				case SteeringEndSwitchAddress:
					setControlRegister(KartControlRegister.SteeringEndSwitchAddress, 
								setup.steeringEndSwitchAddress());
					break;

				case SteeringMaxPosition:
					setSteeringPositionNormalized(0);
					break;
					
				case HardwareSettings:
					setControlRegister(KartControlRegister.HardwareSettings, 
								KartHardwareSettings.toInt(setup.hardwareSettings()));
					break;
			
				case UpdateInterval:
					setControlRegister(KartControlRegister.UpdateInterval,	
									   setup.updateIntervalMs());
					break;
			}
		}
	};
	
	
	/*** Communication ****************************************************************************/
	
	/**
	 * Tries to establish a connection to the given kart entity. The operation is 
	 * 		  asynchronous and returns control direct to the caller. The status of the connection
	 * 		  attempt is returned via the kart listener interface, method 
	 * 		  connectionStatusChanged().
	 * 
	 * @param entity	Kart communication entity to connect to, a such entity can be found using 
	 * 					the method startDiscovery() on the class CommunicationSetup.
	 */
	public void connect(DiscoveryEntity entity) {
		// Call the communications package asynchronous connect method.
		CommunicationSetup.connect(entity, protocolListener, commSetupListener);
	}
	
	/**
	 * Disconnects from the kart.
	 */
	public void disconnect() {
		// Stop updating registers to the kart.
		if(commitTask != null)
			commitTask.stop();
		
		// Close the communication with the kart.
		if (protocol != null)
			protocol.dispose();
		
		// Set protocol to null.
		protocol = null;
	}
	
	// CommunicationSetupListener interface implementation.
	private CommunicationSetupListener commSetupListener = new CommunicationSetupListener() {
		@Override
		public void communicationSetupSucceeded(Protocol protocol) {
			// Connection setup is ok, so save the pointer to the protocol.
			Kart.this.protocol = protocol;
			
			uncommittedControlRegisters = EnumSet.allOf(KartControlRegister.class);
			
			// Start synchronizing "dirty" registers to the kart.
			commitTask = new CommitTask(true);
			commitTask.start();
			
			// Signal new connection status to the kart listeners.
			for (KartListener listener: kartListeners) {
				listener.connectionStatusChanged(Kart.this, true);
			}
		}
		
		@Override
		public void communicationSetupFailed(String reason) {
			// Signal failed connection attempt to all listeners.
			for (KartListener listener: kartListeners) {
				listener.message(Kart.this, "Connection setup error: " + reason);
			}
			
			// Signal new connection status to the kart listeners.
			for (KartListener listener: kartListeners) {
				listener.connectionStatusChanged(Kart.this, false);
			}
		}
	};
	
	// Protocol used to communicate with the kart.
	private Protocol protocol;
	
	// ProtocolListener implementation.
	private ProtocolListener protocolListener= new ProtocolListener() {
		@Override
		public void registerValueIndication(Protocol protocol, KartStatusRegister register, 
											int value) {
			// Update register value in local registers storage.
			statusRegisters[register.getAddress()] = value;
			
			// Notify all listeners about the change.
			for (KartStatusRegisterListener listener: statusRegistersListeners)
				listener.statusRegisterHasChanged(Kart.this, register, value);
		}

		@Override
		public void stateChangeIndication(Protocol protocol, ProtocolState state) {
			// Is the protocol online?
			boolean online = state == ProtocolState.Online;
			
			// Inform all kart listeners about the change.
			for (KartListener listener: kartListeners) {
				listener.connectionStatusChanged(Kart.this, online);
			}
			
			// If we are not online anymore, close the connection.
			if (!online) {
				disconnect();
			}
		}

		@Override
		public void errorIndication(Protocol protocol, String error) {
			// Inform all kart listeners about the error.
			for (KartListener listener: kartListeners) {
				listener.message(Kart.this, "Communication error: " + error);
			}
		}
	};
	
	// Triggers a write of the control registers.
	private void commitControlRegisters() {
		if(commitTask != null)
			commitTask.commitHandler.post(new CommitTask(false));
	}

	// Task that periodically updates the "dirty" registers of the local kart representation to 
	// the real kart.
	private class CommitTask implements Runnable {
		CommitTask(boolean continuously) {
			this.continuously = continuously;
		}
			
		@Override
		public void run() {
			// If the protocol is ready, we commit the registers.
			if (protocol != null && protocol.isReady()) {
				
				// Create a local copy of the set of registers that have to be written to the kart.
				EnumSet<KartControlRegister> uncommitted = uncommittedControlRegisters;
				
				// For each of the registers to commit, write it to the kart.
				for (KartControlRegister register: uncommitted) {
					protocol.setRegisterValue(register, controlRegisters[register.getAddress()]);
					synchronized (uncommittedControlRegisters) {
						uncommittedControlRegisters.remove(register);
					}
				}
			}

			// Reschedule if task is continuously.
			if (continuously)
				commitHandler.postDelayed(this, 100);
		}

		public void start() {
			// Schedule the task for execution ASAP.
			commitHandler.post(this);
		}
		
		public void stop() {
			// Remove the task.
			commitHandler.removeCallbacks(this);
		}
		
		// If true, the task is executed continuously, otherwise only once.
		private boolean continuously;
		public Handler commitHandler = new Handler();
	}
	
	// The task actually updating the kart's registers.
	private CommitTask commitTask;
	
		
	/*** Convenience methods to enable a simple access to the kart. *******************************/
	
	/**
	 * Sets the drive speed (throttle) to the given absolute value. Valid values are 
	 * 		  [-15, 15], whereas negative values mean backward and positive values mean forward.
	 * 
	 * @param speed	Absolute drive speed (throttle).
	 */
	public void setDriveSpeed(int speed) {
		setControlRegister(KartControlRegister.DriveSpeed, 
					KartUtils.limit(speed, MINIMAL_SPEED, MAXIMAL_SPEED));
	}

	/**
	 * Sets the drive speed (throttle) to the given normalized value. Valid values are 
	 * 		  [-1.0, 1.0], whereas negative values mean backward and positive values mean forward.
	 * 
	 * @param speed Normalized drive speed (throttle).
	 */
	public void setDriveSpeedNormalized(float speed) {
		setDriveSpeed((int)(speed * 15));
	}
	
	/**
	 * Increases the drive speed (throttle) by the smallest possible increment.
	 */
	public void increaseDriveSpeed() {
		setDriveSpeed(getDriveSpeed() + 1);
	}
	
	/**
	 * Decreases the drive speed (throttle) by the smallest possible increment.
	 */
	public void decreaseDriveSpeed() {
		setDriveSpeed(getDriveSpeed() - 1);
	}
	
	/**
	 * Returns the actual absolute drive speed (throttle) in the range [-15, 15].
	 * 
	 * @return	Actual absolute drive speed.
	 */
	public int getDriveSpeed() {
		return getControlRegister(KartControlRegister.DriveSpeed);
	}
	
	/**
	 * Returns the actual normalized drive speed (throttle) in the range [-1.0, 1.0].
	 * 
	 * @return	Actual normalized drive speed.
	 */
	public float getDriveSpeedNormalized() {
		return getControlRegister(KartControlRegister.DriveSpeed) / 15f;
	}
	
	/**
	 * Sets the steering position on the kart to the given absolute value. Note that this is
	 * 		  the setpoint of the steering control system on the kart. It may take some time for  
	 * 	      the steering system to actually reach the given position. The position is limited to 
	 * 		  the valid range from 0 to the maximal steering position in the kart setup.
	 * 
	 * @param position	Absolute steering position.
	 */
	public void setSteeringPosition(int position) {
		setControlRegister(KartControlRegister.SteeringPosition, 
					KartUtils.limit(position, 0, setup.steeringMaxPosition()));
	}
	
	/**
	 * Sets the steering position on the kart to the given normalized value. Note that this 
	 * 		  is the setpoint of the steering control system on the kart. It may take some time for  
	 * 	      the steering system to actually reach the given position. The position is limited to 
	 * 		  [-1, 1] whereas -1 and 1 are the left, respective right end positions.
	 * 
	 * @param position	Normalized steering position.
	 */
	public void setSteeringPositionNormalized(float position) {
		setSteeringPosition((int)(position * setup.steeringCenterPosition() + 
							setup.steeringCenterPosition()));
	}
	
	/**
	 * Returns the actual absolute steering control position. This is not the steering 
	 * 		  position the kart actually has, it is the position that the kart is instructed to 
	 * 		  reach.
	 * 
	 * @return	Absolute steering position setpoint.
	 */
	public int getSteeringControlPosition() {
		return getControlRegister(KartControlRegister.SteeringPosition) & 0xFFFF;
	}
	
	/**
	 * Returns the actual normalized steering control position. This is not the steering 
	 * 		  position the kart actually has, it is the position that the kart is instructed to 
	 * 		  reach.
	 * 
	 * @return	Normalized steering position setpoint [-1.0, 1.0].
	 */
	public float getSteeringControlPositionNormalized() {
		return (float)getSteeringControlPosition() / (float)setup.steeringCenterPosition() - 1f;
	}
	
	/**
	 * Turns the steering position by the given absolute amount of steps, but not further
	 * 		  as the left and right end positions. Negative numbers signal opposite direction.
	 * 
	 * @param amount	Absolute number of steps to turn.
	 */
	public void turnSteeringBy(int amount) {
		setSteeringPosition(getSteeringControlPosition() + amount);
	}
	
	/**
	 * Turns the steering position by the given normalized amount, but not further
	 * 		  as the left and right end positions, whereas 1.0 would match the complete steering 
	 * 	      range. Negative numbers signal opposite direction.
	 * 
	 * @param amount	Normalized amount to turn.
	 */
	public void turnSteeringNormalizedBy(float amount) {
		setSteeringPositionNormalized(getSteeringControlPositionNormalized() + 2f * amount);
	}

	/**
	 * Returns the actual absolute steering position (not the setpoint). This is the 
	 * 		  position the steering system actually has and does not necessarily has to match
	 * 		  the setpoint as the system needs some time to reach the requested positions.
	 * 
	 * @return Actual absolute steering position.
	 */
	public int getSteeringPosition() {
		return getStatusRegister(KartStatusRegister.SteeringPosition) & 0xFFFF;
	}
	
	/**
	 * Returns the actual normalized steering position (not the setpoint) in the range 
	 * 		  [-1.0, 1.0]. This is the position the steering system actually has and does not 
	 * 		  necessarily has to match the setpoint as the system needs some time to reach the
	 * 		  requested positions.
	 * 
	 * @return Actual normalized steering position from [-1.0, 1.0].
	 */
	public float getSteeringPositionNormalized() {
		return (float)getSteeringPosition() / (float)setup.steeringCenterPosition() - 1f;
	}

	/**
	 * Returns true if the kart has reached the target steering position, false
	 * 		  otherwise.
	 * 
	 * @return True if the kart has reached the target steering position, false otherwise.
	 */
	public boolean isSteeringPositionReached() {
		return getStatusRegister(KartStatusRegister.SteeringEndSwitchState) == 1;
	}
	
	/**
	 * Returns true if the steering end contact is closed, false otherwise.
	 * 
	 * @return True if steering end contact is closed, false otherwise.
	 */
	public boolean isSteeringEndContactTriggered() {
		return getStatusRegister(KartStatusRegister.SteeringEndSwitchState) == 0;
	}
	
	/**
	 * Returns the battery level from 0.0 (empty) to 1.0 (full).
	 * @return Battery level [0.0, 1.0].
	 */
	public float getBatteryLevel() {
		return KartUtils.limit(((getStatusRegister(
				KartStatusRegister.BatteryVoltageLevel) & 0xFFFF) - MINIMAL_BATT_LEVEL ) / 
				(MAXIMAL_BATT_LEVEL - MINIMAL_BATT_LEVEL), 0f, 1f);
	}
	
	/**
	 * Sets the LED at the give index [0,3] to the given value (true=on).
	 * 
	 * @param index	LED index [0, 3].
	 * @param on	If true the LED is set on, if false the LED is set off.
	 */
	public void setLed(int index, boolean on) {
		if (index < 0 && index >= 4) return;
		int ledMask = getControlRegister(KartControlRegister.LEDs);
		if (on)
			ledMask |= 1 << index;
		else
			ledMask &= ~(1 << index);
		setControlRegister(KartControlRegister.LEDs, ledMask);
	}
	
	/**
	 * Gets the status if the led at the given index.
	 * 
	 * @param index	LED index [0, 3]
	 * @return		True if the LED is on, false if the LED is off.
	 */
	public boolean getLed(int index) {
		if (index < 0 && index >= 4) return false;
		int ledMask = getControlRegister(KartControlRegister.LEDs);
		return (ledMask & (1 << index)) != 0;	
	}
	
	/**
	 * Toggles the LED at the given index.
	 * 
	 * @param index LED index [0, 3]
	 */
	public void toggleLed(int index) {
		if (index < 0 && index >= 4) return;
		int ledMask = getControlRegister(KartControlRegister.LEDs);
		ledMask ^= 1 << index;
		setControlRegister(KartControlRegister.LEDs, ledMask);
	}
	
	
	/*** Control register API *********************************************************************/
	
	/**
	 * Sets the given control register to the given value. Note that control registers hold
	 * 		  only 16 bit values, so not the whole range of integer values is applicable.
	 * 
	 * @see KartControlRegister
	 * @param register	The register to set.
	 * @param value		The value to set for the register.
	 */
	public void setControlRegister(KartControlRegister register, int value) {
		
		boolean changed = value != controlRegisters[register.getAddress()];
		controlRegisters[register.getAddress()] = value;
		if (changed) {
			synchronized (uncommittedControlRegisters) {
				uncommittedControlRegisters.add(register);
				commitControlRegisters();
			}
		}
	}
	
	/**
	 * Returns the actual value of the control register. The value is casted signed to an
	 * 		  int, so if you like to read the register value rather unsigned, use the 
	 * 		  getControlRegisterUnsigned() method.
	 * 
	 * @param register	The control register to read.
	 * @return			The actual SIGNED register value.
	 */
	public int getControlRegister(KartControlRegister register) {
		return controlRegisters[register.getAddress()];
	}

	/**
	 * Returns the actual value of the control register. The value is casted unsigned to an
	 * 		  int.
	 * 
	 * @param register	The control register to read.
	 * @return			The actual UNSIGNED register value.
	 */
	public int getControlRegisterUnsigned(KartControlRegister register) {
		return controlRegisters[register.getAddress()] & 0xFFFF;
	}

	// Control register storage.
	private int[] controlRegisters = new int[KartControlRegister.COUNT];
	
	// "Dirty" registers that have to be sent to the kart.
	private EnumSet<KartControlRegister> uncommittedControlRegisters = 
			EnumSet.allOf(KartControlRegister.class);
	
	
	/*** Status parameters ************************************************************************/
	
	/**
	 * Returns the actual value of the status register. The value is casted signed to an
	 * 		  int, so if you like to read the register value rather unsigned, use the 
	 * 		  getStatusRegisterUnsigned() method.
	 * 
	 * @param register	The status register to read.
	 * @return			The actual SIGNED register value.
	 */
	public int getStatusRegister(KartStatusRegister register) {
		return statusRegisters[register.getAddress()];
	}
	
	/**
	 * Returns the actual value of the status register. The value is casted unsigned to an
	 * 		  int.
	 * 
	 * @param register	The status register to read.
	 * @return			The actual UNSIGNED register value.
	 */
	public int getStatusRegisterUnsigned(KartStatusRegister register) {
		return statusRegisters[register.getAddress()] & 0xFFFF;
	}

	// Status register storage.
	private int[] statusRegisters = new int[KartStatusRegister.COUNT];
	
	
	/*** Status register listeners - low level interface ******************************************/
	
	/**
	 * Adds the given status listener to the kart. The listener will be notified about the 
	 * 		  change of any status register on the kart.
	 * 
	 * @param listener	Reference to the listener to add.
	 */
	public void addStatusRegisterListener(KartStatusRegisterListener listener) {
		statusRegistersListeners.add(listener);
	}
	
	/**
	 * Removes the given status register listener.
	 * 
	 * @param listener Reference to the listener to remove.
	 */
	public void removeStatusRegisterListener(KartStatusRegisterListener listener) {
		statusRegistersListeners.remove(listener);
	}
	
	// List of status register listeners.
	private List<KartStatusRegisterListener> statusRegistersListeners = 
			new LinkedList<KartStatusRegisterListener>();
	
	
	/*** Kart listeners - high level (limited) interface ******************************************/
	
	/**
	 * Adds the given kart listener to the kart.
	 * 
	 * @param listener	Reference to the object implementing the kart listener interface.
	 */
	public void addKartListener(KartListener listener) {
		if (listener != null && !kartListeners.contains(listener)) 
			kartListeners.add(listener);
	}
	
	/**
	 * Removes the kart listener.
	 * 
	 * @param listener Reference to the object implementing the kart listener interface to remove.
	 */
	public void removeKartListener(KartListener listener) {
		if (listener != null) kartListeners.remove(listener);
	}
	
	// List of all kart listeners.
	private List<KartListener> kartListeners = new LinkedList<KartListener>();
	
	// A status register implementation that notifies the Kart Listeners upon important events.
	private KartStatusRegisterListener kartListenerDispatcher = new KartStatusRegisterListener() {
		@Override
		public void statusRegisterHasChanged(Kart kart, KartStatusRegister register, 
											 int newValue) {
			switch (register) {
				case SteeringPosition:
					for (KartListener listener: kartListeners)
						listener.steeringPositionChanged(Kart.this, getSteeringPosition(),
								getSteeringPositionNormalized());
					break;
					
				case SteeringStatus:
					for (KartListener listener: kartListeners)
						listener.steeringPositionHasReached(Kart.this, 
								isSteeringPositionReached());
					break;
					
				case BatteryVoltageLevel:
					for (KartListener listener: kartListeners)
						listener.batteryVoltageChanged(Kart.this, getBatteryLevel());
					break;
					
				default:
					break;
			}
		}
	};
}
