package ch.hei.kart;

/**
 * Enumeration in order to identify the different control registers of the Kart.
 * 
 * @author Michael Clausen (clm@hevs.ch)
 */
public enum KartControlRegister {
	/**
	 * The Drive PWM period as UINT16. Divider value from the FPGA main clock  
	 * 		  frequency to the PWM step. Default value is 0x40.
	 */
	DrivePwmPeriod(0),
	
	/**
	 * Drive motor speed as INT5. Negative numbers = backwards, positive numbers = forwards.
	 */
	DriveSpeed(1),
	
	/**
	 * The Stepper (Steering) motor step period as UINT16. Divider value from the clock 
	 * 		  frequency to the stepper motor step period. Default value is 0x100.
	 */
	SteeringStepPeriod(2),
	
	/**
	 * Steering position setpoint as UINT16 register. Stepper motor target position (zero is 
	 * 		  stepper end sensor, values are positive only).
	 */
	SteeringPosition(3),
	
	/**
	 * Specifies the address at which the end switch is read at bit 0 (if it isn't local).
	 */
	SteeringEndSwitchAddress(4),
	
	/**
	 * Hardware settings register. Defines how the motor rotations are linked to the kart 
	 * 		  move actions.
	 * 
	 * Bit 0: Kart goes as DC motor.
	 *        When '0', the kart goes backwards when the motor turns forward.
	 *        
	 * Bit 1: Kart turns right (clockwise).
	 * 		  When '1', kart turns to the right as stepper motor phases go from 1 to 4.
	 * 
	 * Bit 2: End sensor is on left side.
	 * 		  If '1', the angles are measured clockwise.
	 * 
	 * Bit 3: Stepper end emulation.
	 *        Emulates an end switch contact for the direction motor.
	 *        
	 * Bit 4: Restart slaves.
	 * 	      Puts following slaves back in startup mode: stepper motor
	 */
	HardwareSettings(5),
	
	/**
	 * LED control register. Allows to change the state of the 4 onboard LED drivers.
	 */
	LEDs(6),
	
	/**
	 * Update interval (from kart to App) in milliseconds.
	 */
	UpdateInterval(15);
	
	private KartControlRegister(int address) {
		this.address = address;
	}
	
	/**
	 * Returns the address of the register in the kart's memory.
	 * 
	 * @return Address of the register.
	 */
	public int getAddress() {
		return address;
	}

	// The address of the register.
	private int address;
	
	// Maximal count of registers.
	static final int COUNT = 16;
}
