package ch.hei.kart;

import java.util.EnumSet;

import ch.hei.kart.KartSetupListener.SetupParameter;

/**
 * Groups all kart settings together and allows fancy Kart setup using returning reference
 * 		  listening and setter chaining techniques.
 * 
 * @author Michael Clausen (clm@hevs.ch)
 */
public class KartSetup {
	// Creates a new kart setup object with the default values.
	KartSetup() {
		drivePwmPeriod = 64;
		steeringMaxPosition = 1000;
		steeringStepPeriod = 256;
		steeringEndSwitchAddress = 4;
		hardwareSettings = EnumSet.noneOf(KartHardwareSettings.class);
		updateInterval = 64;
	}
	
	/**
	 * Returns the drive motor control block PWM period. The period is defined as the value the 
	 * main FPGA clock is divided by.
	 * 
	 * @return PWM period.
	 */
	public int drivePwmPeriod() {
		return drivePwmPeriod;
	}
	
	/**
	 * Sets the drive motor control block PWM period. The period is defined as the value the 
	 * main FPGA clock is divided by. Defaults to 64.
	 * 
	 * @param period	PWM period.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup drivePwmPeriod(int period) {
		drivePwmPeriod = KartUtils.limit(period, 1, 65535);
		if (listener != null) listener.parameterChanged(SetupParameter.DrivePwmPeriod);
		return this;
	}
	
	/**
	 * Returns the maximal step count at the opposite site of the stepper end contact switch.
	 * 
	 * @return Maximal stepper motor step count.
	 */
	public int steeringMaxPosition() {
		return steeringMaxPosition;
	}

	// Returns the stepper motor center position.
	int steeringCenterPosition() {
		return steeringMaxPosition / 2;
	}
	
	/**
	 * Sets the maximal step count of the steering motor. Default value is defined by Kart VHDL 
	 * design.
	 * 
	 * @param position	The value of the step counter at the opposite site of the end switch.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup steeringMaxPosition(int position) {
		steeringMaxPosition = KartUtils.limit(position, 0, 65535);
		if (listener != null) listener.parameterChanged(SetupParameter.SteeringMaxPosition);
		return this;
	}
	
	/**
	 * Returns the steering motor driver step period. The period is defined as the value the 
	 * main FPGA clock is divided by.
	 * 
	 * @return Steering motor step period.
	 */
	public int steeringStepPeriod() {
		return steeringStepPeriod;
	}
	
	/**
	 * Sets the steering motor driver step period. The period is defined as the value the 
	 * main FPGA clock is divided by. Defaults to 256.
	 * 
	 * @param period 	Steering motor step period.
	 * @return		 	Reference to itself in order to chain the setters.
	 */
	public KartSetup steeringStepPeriod(int period) {
		steeringStepPeriod = KartUtils.limit(period, 1, 65535);
		if (listener != null) listener.parameterChanged(SetupParameter.SteeringStepPeriod);
		return this;
	}
	
	/**
	 * Returns the steering end switch address from which emulated end switch the data is 
	 * considered by the steering FSM. This value can be ignored if the option to enable the
	 * emulated end contact is not set in the hardware settings.
	 * 
	 * @return Steering end contact address.
	 */
	public int steeringEndSwitchAddress() {
		return steeringEndSwitchAddress;
	}
	
	/**
	 * Sets the steering end switch address from which emulated end switch the data is 
	 * considered by the steering FSM. This value will be ignored if the option to enable the
	 * emulated end contact is not set in the hardware settings. Defaults to 0.
	 * 
	 * @param address	Steering end contact address.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup steeringEndSwitchAddress(int address) {
		steeringEndSwitchAddress = KartUtils.limit(address, 0, 31);
		if (listener != null) listener.parameterChanged(SetupParameter.SteeringEndSwitchAddress);
		return this;
	}
	
	/**
	 * Returns the Kart's actual hardware settings. Default is all flags cleared.
	 * 
	 * @return Kart's hardware settings.
	 */
	public EnumSet<KartHardwareSettings> hardwareSettings() {
		return hardwareSettings;
	}
	
	/**
	 * Sets the given hardware setting flag.
	 * 
	 * @param setting	Hardware setting flag to set.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup setHwSetting(KartHardwareSettings setting) {
		hardwareSettings.add(setting);
		if (listener != null) listener.parameterChanged(SetupParameter.HardwareSettings);
		return this;
	}
	
	/**
	 * Clears the given hardware setting flag.
	 * 
	 * @param setting	Hardware setting flag to clear.
	 * @return			Reference to itself in order to chain the setters.
	 */
	public KartSetup clearHwSetting(KartHardwareSettings setting) {
		hardwareSettings.remove(setting);
		if (listener != null) listener.parameterChanged(SetupParameter.HardwareSettings);
		return this;
	}
	
	/**
	 * Returns the update interval of the Kart in seconds. This interval is the time between each
	 * register update send from the Kart to the Android application.
	 * 
	 * @return Update interval in seconds.
	 */
	public float updateInterval() {
		return (float)updateInterval / 1000f;
	}
	
	// Returns the update interval in milliseconds.
	int updateIntervalMs() {
		return updateInterval;
	}
	
	/**
	 * Sets the update interval of the Kart in seconds. This interval is the time between each
	 * register update send from the Kart to the Android application. Note that fractions of 
	 * seconds is supported too. Defaults to 0.064.
	 * 
	 * @param interval	Update interval in seconds.
	 * @return
	 */
	public KartSetup updateInterval(float interval) {
		updateInterval = (int)(KartUtils.limit(interval, 0.001f, 65f) * 1000f);
		if (listener != null) listener.parameterChanged(SetupParameter.UpdateInterval);
		return this;
	}
	
	// The attributes containing the actual setup...
	private int drivePwmPeriod;
	private int steeringMaxPosition;
	private int steeringStepPeriod;
	private int steeringEndSwitchAddress;
	private EnumSet<KartHardwareSettings> hardwareSettings;
	private int updateInterval;
	
	// Kart listens to changes on his setup...
	void setListener(KartSetupListener listener) {
		this.listener = listener;
	}

	private KartSetupListener listener = null;
}
