package ch.hei.kart.comm.discovery;

import android.os.Handler;
import android.os.Looper;
import android.os.Message;

/**
 * Wrapper to ensure that the callbacks from the discovery are always called inside the 
 * 		  application's main thread (Android is very fussy about that).
 * 
 * @author Michael Clausen (clm@hevs.ch)
 */
public class UiThreadDiscoveryListenerWrapper implements DiscoveryListener {
	/**
	 * Sets the actual listener wrapped by the thread wrapper.
	 * 
	 * @param listener The listener to set up.
	 */
	public void setListener(DiscoveryListener listener) {
		this.listener = listener;
	}
	
	// The type of the message (We send messages between threads to transfer the callbacks).
	private enum MessageType {
		EntityFoundIndication,
		ErrorIndication
	}
	
	// The actual message content we send between the threads.
	private class MessageContent {
		// Creates a new message for the entityFoundIndication() method.
		public MessageContent(Discovery discovery, DiscoveryEntity entity, 
							  DiscoveryListener listener) {
			this.type = MessageType.EntityFoundIndication;
			this.subject = entity;
			this.discovery = discovery;
			this.listener = listener;
		}
		
		// Creates a new message for the errorIndication() method.
		public MessageContent(Discovery discovery, String errorMessage, 
							  DiscoveryListener listener) {
			this.type = MessageType.ErrorIndication;
			this.subject = errorMessage;
			this.discovery = discovery;
			this.listener = listener;
		}
		
		public MessageType type;
		public Object subject;
		public Discovery discovery;
		public DiscoveryListener listener;
	}
	
	@Override
	public void entityFoundIndication(Discovery discovery, DiscoveryEntity entity) {
		// If listener is empty, we have nothing to do.
		if (listener == null) return;
		
		// If we are actually in the main thread, we call the listener method directly.
		if (Looper.getMainLooper().getThread() == Thread.currentThread()) {
			listener.entityFoundIndication(discovery, entity);
		} else {
			// Otherwise we send a message to the main loop that will call the listener later.
			Message message = Message.obtain();
			message.obj = new MessageContent(discovery, entity, listener);
			handler.sendMessage(message);
		}
	}

	@Override
	public void errorIndication(Discovery discovery, String error) {
		// If listener is empty, we have nothing to do.
		if (listener == null) return;
		
		// If we are actually in the main thread, we call the listener method directly.
		if (Looper.getMainLooper().getThread() == Thread.currentThread()) {
			listener.errorIndication(discovery, error);
		} else {
			// Otherwise we send a message to the main loop that will call the listener later.
			Message message = Message.obtain();
			message.obj = new MessageContent(discovery, error, listener);
			handler.sendMessage(message);
		}
	}
	
	// The real listener wrapped by this object.
	private DiscoveryListener listener;
	
	// Main thread handler that will handle the messages posted by this wrapper.
	private static Handler handler = new Handler() {
		@Override
		public void handleMessage(Message message) {
			// Get message content and switch depending the message type.
			MessageContent content = (MessageContent)message.obj;
			switch (content.type) {	
				case EntityFoundIndication:
					// Call the entityFoundIndication() on the real listener.
					if (content.listener != null)
						content.listener.entityFoundIndication(content.discovery, 
															   (DiscoveryEntity)content.subject);
					break;
					
				case ErrorIndication:
					// Call the entityFoundIndication() on the real listener.
					if (content.listener != null)
						content.listener.errorIndication(content.discovery, 
														 (String)content.subject);
					break;
			}
		}
	};
}
