package ch.hei.kart.comm.transport;

import java.nio.ByteBuffer;

import ch.hei.kart.comm.discovery.DiscoveryEntity;

/**
 * Interface for a Kart communication transport layer. A Kart can be controlled using 
 * 		  different transport layers like Bluetooth or TCP. This is the common interface they
 * 		  have to provide.
 * 
 * @author Michael Clausen (clm@hevs.ch)
 */
public interface Transport {
	/**
	 * Tries to establish a connection to the endpoint described by the connection string.
	 * 
	 * @param connectionString	The connection string of the object to connect to.
	 * @return					True on success, false otherwise.
	 */
	public void connect(String connectionString) throws Exception;
	
	/**
	 * Tries to establish a connection to the endpoint described by the entity.
	 * 
	 * @param entity	The entity to connect to.
	 * @return			True on success, false otherwise.
	 */
	public void connect(DiscoveryEntity entity) throws Exception;
	
	/**
	 * Returns true if the transport layer actually is connected, false otherwise.
	 *
	 * @return	True if ready, false otherwise.
	 */
	public boolean isConnected();
	
	/**
	 * Disconnects the communication.
	 */
	public void disconnect();
	
	/**
	 * Sets the listener that will get the transport layer events.
	 * 
	 * @param listener	Reference to the listener to set.
	 */
	public void setListener(TransportListener listener);
	
	/**
	 * Writes the given byte buffer to the peer.
	 * 
	 * @param buffer	Buffer to write.
	 * @return			True if the send process could be started in the background or the data was
	 * 					send, false on error.
	 */
	public boolean write(ByteBuffer buffer);
}
