package ch.hei.kart.comm.transport.bluetooth;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.nio.ByteBuffer;
import java.util.UUID;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;

import ch.hei.kart.comm.discovery.DiscoveryEntity;
import ch.hei.kart.comm.transport.Transport;
import ch.hei.kart.comm.transport.TransportListener;
import ch.hei.kart.comm.transport.TransportState;

public class RfCommTransport implements Transport {
	@Override
	public void connect(String connectionString) throws Exception {

		// Ensure that the adapter is present and enabled.
		if (btAdapter == null) btAdapter = BluetoothAdapter.getDefaultAdapter();
		if (btAdapter == null || !btAdapter.isEnabled())
			throw new IOException("Bluetooth adatper is deactivated, please enable it to " + 
								  "find Bluetooth devices.");
		
		// Try to get the device.
		btDevice = btAdapter.getRemoteDevice(connectionString);
		if (btDevice == null ) throw new NullPointerException("Could not create a Bluetooth " +
															  "device!");
		
		// If the device is not bound, try to bind it...
		if (btDevice.getBondState() != BluetoothDevice.BOND_BONDED) {
			// Hack. Use reflection to bond the device...
			Class<?> bluetoothDeviceClass = Class.forName("android.bluetooth.BluetoothDevice");
			Method createBondMethod = bluetoothDeviceClass.getMethod("createBond");
			Boolean returnValue = (Boolean)createBondMethod.invoke(btDevice);
			if (!returnValue)
				throw new IOException("Unable to pair with device!");
			while(btDevice.getBondState() != BluetoothDevice.BOND_BONDED) {
				Thread.sleep(250);
			}
		}
		if (btDevice.getBondState() != BluetoothDevice.BOND_BONDED) {
			btDevice = null;
			throw new IOException("Unable to pair with device!");
		}
		
		// Try to establish the RFCOMM connection.
		btSocket = btDevice.createRfcommSocketToServiceRecord(
				UUID.fromString("00001101-0000-1000-8000-00805F9B34FB"));
		btSocket.connect();
		btInput = btSocket.getInputStream();
		btOutput = btSocket.getOutputStream();
	
		btReader.start();		
	}

	@Override
	public void connect(DiscoveryEntity entity) throws Exception {
		connect(((RfCommDiscoveryEntity)entity).getDeviceAddress());
	}

	@Override
	public boolean isConnected() {
		return btSocket != null && btSocket.isConnected();
	}
	
	@Override
	public void disconnect() {
		if (btSocket == null) return;
		
		btInput = null;
		btOutput = null;
		if (btSocket != null) {
			try {
				btSocket.close();
			} catch (IOException exception) {
				exception.printStackTrace();
			}
			btSocket = null;
		}
		btDevice = null;
		listener = null;
		
		setState(TransportState.ConnectionLost);
	}
	
	@Override
	public void setListener(TransportListener listener) {
		this.listener = listener;
	}

	@Override
	public boolean write(ByteBuffer buffer) {
		if (btSocket == null || btOutput == null || ! btSocket.isConnected()) return false;
		
		try {
			buffer.flip();
			byte[] data = new byte[buffer.remaining()];
			buffer.get(data);
			
			btOutput.write(data);
			return true;
		} catch (Exception exception) {
			exception.printStackTrace();
			disconnect();
			return false;
		}
	}

	private BluetoothAdapter btAdapter;
	private BluetoothDevice btDevice;
	private BluetoothSocket btSocket;
	private InputStream btInput;
	private OutputStream btOutput;
	private Thread btReader = new Thread() {
		@Override
		public void run() {
			byte[] buffer = new byte[128];
			while(btInput != null) {
				try {
					int size = btInput.read(buffer);
					ByteBuffer byteBuffer = ByteBuffer.allocate(size);
					byteBuffer.put(buffer, 0, size);
					byteBuffer.flip();
					if (listener != null) listener.dataIndication(RfCommTransport.this, byteBuffer);
				} catch (IOException exception) {
					if (btInput != null) {
						if (listener != null) listener.errorIndication(RfCommTransport.this, 
							   	   exception.getMessage());
						exception.printStackTrace();
					}
					disconnect();
				}
			}
		}
	};
	private TransportListener listener = null;
	
	private TransportState state;
	
	private void setState(TransportState state) {
		if (state != this.state) {
			this.state = state;
			if (listener != null) listener.stateChangeIndication(this, state);
		}
	}
}
